// -Pixel Plotting Demo-
// Demonstrates DirectDraw
// Written by J. Farrell

// This program demonstrates pixel plotting in a palettized mode,
// 640x480x8. When you press Escape, it switches to 16-bit RGB
// mode, 640x480x16. Press Escape again to end the demo.

#define WIN32_LEAN_AND_MEAN

// To compile this program, remember what you need to do to set up
// Visual C++ for DirectX. Make sure the DirectX SDK directories
// for libraries and headers are FIRST on their repective lists
// (under Tools menu, go to "Options..."). Then create a new project
// using this as the only source file, and remember to link the
// ddraw.lib and dxguid.lib libraries to your project!

// includes
#include <windows.h>
#include <windowsx.h>
#include <ddraw.h>

// macros
#define INIT_DXSTRUCT(dxstruct) { ZeroMemory(&dxstruct, sizeof(dxstruct)); dxstruct.dwSize = sizeof(dxstruct); }
#define KEYSTATE(key) ((GetAsyncKeyState(key) & 0x8000) ? TRUE : FALSE)
#define RGB_16BIT555(r, g, b) ((r<<10) | (g<<5) | (b))
#define RGB_16BIT565(r, g, b) ((r<<11) | (g<<5) | (b))

// G L O B A L S /////////////////////////////////////////////////////
// Windows
HINSTANCE hMainInstance;
HWND hMainWindow;

// DirectDraw
LPDIRECTDRAW7 lpdd7 = NULL;
LPDIRECTDRAWSURFACE7 lpddsPrimary = NULL;
LPDIRECTDRAWPALETTE lpddpal = NULL;
DDSURFACEDESC2 ddsd;

// data
int nCurrentColorDepth = 8; // are we palettized or RGB?
int nPixelFormat; // we'll set this to 15 for 555, or 16 for 565
PALETTEENTRY palette[256]; // our color lookup table

//////////////////////////////////////////////////////////////////////

// prototypes
LRESULT CALLBACK WindowProc(HWND, UINT, WPARAM, LPARAM);
int Initialize(void);
inline void PlotPixel16(int x, int y, USHORT color, USHORT* buffer, int nPitch);
inline void PlotPixel8(int x, int y, UCHAR color, UCHAR* buffer, int byPitch);
inline void PlotPixelFast16(int x, int y, USHORT color, USHORT* buffer);
void Shutdown(void);
int SwitchColorDepth(void);

//////////////////////////////////////////////////////////////////////
// WINMAIN
// Program entry point. Handles the window setup and the contains the
// program's main loop.
//////////////////////////////////////////////////////////////////////
int APIENTRY WinMain(HINSTANCE hinstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	WNDCLASSEX winclass; // our window class
	HWND hwnd;           // handle to a window
	MSG msg;             // a message variable
	int x;               // loop counter
	
	// fill out the winclass structure
	winclass.cbSize =        sizeof(WNDCLASSEX);
	winclass.style =         CS_DBLCLKS | CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	winclass.lpfnWndProc =   WindowProc;        
	winclass.cbClsExtra =    0;                  // extra class info
	winclass.cbWndExtra =    0;                  // extra window info
	winclass.hInstance =     hinstance;	         
	winclass.hIcon =         LoadIcon(NULL, IDI_WINLOGO);
	winclass.hCursor =       LoadCursor(NULL, IDC_ARROW);
	winclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winclass.lpszMenuName =  NULL;
	winclass.lpszClassName = "TME Class";  
	winclass.hIconSm =       LoadIcon(NULL, IDI_WINLOGO);
	
	// register class
	if (!RegisterClassEx(&winclass))
		return(0);
	
	// create a window
	if (!(hwnd = CreateWindowEx(NULL,            // extended style
			     "TME Class",                    // class identifier
				 "Terran Map Editor",            // window title
				 WS_POPUP | WS_VISIBLE,          // you can't resize me! hahaha!
				 0, 0,							 // initial position
				 640, 480,						 // initial size
				 NULL,                           // handle to parent
				 NULL,
				 hinstance,                      // app handle
				 NULL)))                         // extra parameters
		return(0);

	// set globals
	hMainWindow = hwnd;
	hMainInstance = hinstance;

	// initialize, test for success
	if (Initialize())
	{
		// set up text color and background

		// main program loop
		while (TRUE)
		{
			// check the message queue
			if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if (msg.message == WM_QUIT) // exit loop on quit
					break;
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			// lock the surface
			INIT_DXSTRUCT(ddsd);
			if (FAILED(lpddsPrimary->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL)))
			{
				// end the demo if Lock() fails
				MessageBox(hMainWindow, "Surface lock failed!", "Error", MB_OK | MB_ICONEXCLAMATION);
				break;
			}
			
			// Now look at this. After all that talk about making
			// your code fast, I use thousands upon thousands of
			// random numbers and modulus operations. :) Not the
			// fastest things in the world! But hey, it's just a
			// demo to show you how this stuff works.

			// plot a thousand pixels at random
			if (nCurrentColorDepth == 8)
			{
				// palettized pixel plot
				UCHAR* buffer = (UCHAR*)ddsd.lpSurface;
				int byPitch = ddsd.lPitch;

				for (x=0; x<1000; x++)
					PlotPixel8(rand()%640, rand()%480, rand()%256, buffer, byPitch);
			}
			else
			{
				// RGB pixel plot
				USHORT* buffer = (USHORT*)ddsd.lpSurface;
				int nPitch = (ddsd.lPitch>>1); // divide by 2 to get number of USHORTs
				USHORT color;

				// check color depth
				if (nPixelFormat == 15)
					color = RGB_16BIT555(rand()%32, rand()%32, rand()%32);
				else
					color = RGB_16BIT565(rand()%32, rand()%64, rand()%32);
				
				// now check pitch
				if (nPitch == 640)
				{
					// pitch is 640 - use the fast function
					for (x=0; x<1000; x++)
						PlotPixelFast16(rand()%640, rand()%480, color, buffer);
				}
				else
				{
					// pitch is not 640 - you've got a weird computer :)
					for (x=0; x<1000; x++)
						PlotPixel16(rand()%640, rand()%480, color, buffer, nPitch);
				}
			}

			// unlock surface
			lpddsPrimary->Unlock(NULL);

			// check escape key
			if (KEYSTATE(VK_ESCAPE))
			{
				if (nCurrentColorDepth == 8)
				{
					// switch to RGB mode
					if (!SwitchColorDepth())
					{
						MessageBox(hMainWindow, "Color depth switch failed!", "Error", MB_OK | MB_ICONEXCLAMATION);
						break;
					}
				}
				else
				{
					// quit the demo
					PostQuitMessage(0);
				}
			}
		}
	}
	else
		MessageBox(hMainWindow, "Initialization failed!", "Error", MB_OK | MB_ICONEXCLAMATION);
	
	// shut everything down
	Shutdown();
	MessageBox(NULL, "Shutdown OK!", "Notice", MB_OK | MB_ICONINFORMATION);

	// give Windows the shutdown code
	return(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg,
							WPARAM wparam, LPARAM lparam)
{
	// call the default message handler for everything!
	// am I lazy or what? :)
	return(DefWindowProc(hwnd, msg, wparam, lparam));
}

///////////////////////////////////////////////////////////////
// INITIALIZE
// Creates all the DirectX objects we'll need for this demo,
// sets the screen resolution to 640x480x8.
///////////////////////////////////////////////////////////////
int Initialize(void)
{
	int x;
	
	// initialize COM
	if (FAILED(CoInitialize(NULL)))
		return(FALSE);

	// create DirectDraw object
	if (FAILED(CoCreateInstance(CLSID_DirectDraw, NULL, CLSCTX_ALL,
								IID_IDirectDraw7, (void**)&lpdd7)))
		return(FALSE);
	
	// initialize DirectDraw object
	if (FAILED(lpdd7->Initialize(NULL)))
		return(FALSE);

	// set cooperative level to fullscreen
	if (FAILED(lpdd7->SetCooperativeLevel(hMainWindow, DDSCL_FULLSCREEN | 
		                                               DDSCL_EXCLUSIVE |
													   DDSCL_ALLOWREBOOT)))
		return(FALSE);

	// hide the mouse
	ShowCursor(FALSE);

	// set 640x480x8 palettized mode
	if (FAILED(lpdd7->SetDisplayMode(640, 480, 8, 0, 0)))
		return(FALSE);
	MessageBox(hMainWindow, "Screen mode OK!", "Notice", MB_OK | MB_ICONINFORMATION);

	// create our primary surface
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsPrimary, NULL)))
		return(FALSE);
	MessageBox(hMainWindow, "Primary surface OK!", "Notice", MB_OK | MB_ICONINFORMATION);

	// set a random palette - I'm leaving out the first and last
	// entries, and going to let them default to black and white
	srand(GetTickCount());
	for (x=1; x<255; x++)
	{
		palette[x].peRed = rand() & 0xFF;
		palette[x].peBlue = rand() & 0xFF;
		palette[x].peGreen = rand() & 0xFF;
		palette[x].peFlags = PC_NOCOLLAPSE;
	}

	// now create and initialize our palette object
	if (FAILED(lpdd7->CreatePalette(DDPCAPS_8BIT | DDPCAPS_INITIALIZE,
		                            palette, &lpddpal, NULL)))
		return(FALSE);
	MessageBox(hMainWindow, "Palette OK!", "Notice", MB_OK | MB_ICONINFORMATION);

	// attach the palette to the primary surface
	if (FAILED(lpddsPrimary->SetPalette(lpddpal)))
		return(FALSE);

	// we're ready to rock
	MessageBox(hMainWindow, "Initialization OK!\nYou like the randomly-colored message box? :)", "Notice", MB_OK | MB_ICONINFORMATION);
	return(TRUE);
}

///////////////////////////////////////////////////////////////
// PLOTPIXEL16
// PLOTPIXEL8
// PLOTPIXELFAST16
// Our various pixel-plotting functions. The one labeled 'fast'
// is set to use the fact that the pitch is 640 USHORTs. Remember, 
// this will not always be the case. We test for it in the 
// SwitchColorDepth() function.
///////////////////////////////////////////////////////////////
inline void PlotPixel16(int x, int y, USHORT color, USHORT* buffer, int nPitch)
{
	buffer[y * nPitch + x] = color;
}

inline void PlotPixel8(int x, int y, UCHAR color, UCHAR* buffer, int byPitch)
{
	buffer[y * byPitch + x] = color;
}

inline void PlotPixelFast16(int x, int y, USHORT color, USHORT* buffer)
{
	buffer[(y<<9) + (y<<7) + x] = color;
}

///////////////////////////////////////////////////////////////
// SHUTDOWN
// Release the objects we created.
///////////////////////////////////////////////////////////////
void Shutdown(void)
{
	if (lpddsPrimary)
	{
		// kill the primary surface!
		lpddsPrimary->Release();
		lpddsPrimary = NULL;
	}

	if (lpdd7)
	{
		// blow away DirectDraw!
		lpdd7->Release();
		lpdd7 = NULL;
	}

	// destroy the window!
	CloseWindow(hMainWindow);

	// assassinate COM! :)
	CoUninitialize();
}

///////////////////////////////////////////////////////////////
// SWITCHCOLORDEPTH
// Gets rid of the palette, releases the old primary surface,
// switches the resolution to 16-bit color, and creates a new
// primary surface to match this new screen mode.
///////////////////////////////////////////////////////////////
int SwitchColorDepth(void)
{
	DDPIXELFORMAT pixel;

	// first kill the palette
	lpddpal->Release();
	lpddpal = NULL;

	// then the primary surface
	lpddsPrimary->Release();
	lpddsPrimary = NULL;

	// now switch the screen mode
	if (FAILED(lpdd7->SetDisplayMode(640, 480, 16, 0, 0)))
		return(FALSE);

	// recreate the primary surface
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsPrimary, NULL)))
		return(FALSE);

	// now query the surface to find the pixel format
	INIT_DXSTRUCT(pixel);
	lpddsPrimary->GetPixelFormat(&pixel);
	if (pixel.dwGBitMask == 0x07E0)
		nPixelFormat = 16;
	else
		nPixelFormat = 15;

	// set our color depth variable
	nCurrentColorDepth = 16;

	// all set
	return(TRUE);
}